# -*- coding: utf-8 -*-
"""
Linear hydraulic simulator
Created on Fri Jan 23 11:27:02 2018

@author: Peter Nachtwey
         President
         Delta Computer System, Inc.

"""
from math import pi
import numpy as np
from scipy.integrate import odeint
import matplotlib.pyplot as plt

DT = 0.001                   # update time


class Model:
    """ Defines a linearized hydraulic cylinder """
    def __init__(self, gain, damping_factor, natural_frequency):
        self._gain = gain               # open loop gain (mm/s)/%
        self._df = damping_factor       # normally about 0.3333333
        self._nf = natural_frequency    # radians per second

    def update(self, pva, time_interval, control):
        """ Linearized model for a hydraulic actuator """
        _vel = pva[1]                   # pva has the last
        _acc = pva[2]                   # position, velocity and acceleration    
        _dposdt = _vel
        _dveldt = _acc
        _daccdt = (-self._nf*self._nf*_vel
                   - 2.0*self._df*self._nf*_acc
                   + self._nf*self._nf*self._gain*control)
        _dpvadt = [_dposdt, _dveldt, _daccdt]     # return the rate of change
        return _dpvadt


class ClosedLoop:
    """ The closed loop class can implement P, PI or PID control """
    def __init__(self, kp, ki, kd):
        self._kp = kp                   # proportional gain
        self._ki = ki                   # integrator gain
        self._kd = kd                   # derivative gain
        self._int = 0.0                 # integrator % output
        self._last_error = 0.0          # last position error

    def P(self, r, x):                  # proportional only control
        _tar_pos = r[0]                 # target position
        _act_pos = x[0]                 # actual posiiton
        _error = _tar_pos - _act_pos    # position error
        _control = max(min(self._kp*_error, 100.0), -100.0)
        return _control                 # control output %

    def PI(self, r, x):                 # proportional with integrator
        _tar_pos = r[0]                 # target position
        _act_pos = x[0]                 # actual position
        _error = _tar_pos - _act_pos    # position error
        self._int = max(min(self._int+self._ki*_error*DT, 100.0), -100.0)
        _control = max(min(self._int+self._kp*_error, 100.0), -100.0)
        return _control                 # control output %

    def PID(self, r, x):                # proportional, integral, derivatitive
        _tar_pos = r[0]                 # target position
        _act_pos = x[0]                 # actual position
        _error = _tar_pos - _act_pos    # position error
        self._int = max(min(self._int+self._ki*_error*DT, 100.0), -100.0)
        _control = max(min(self._int+self._kp*_error
                + self._kd*(_error-self._last_error)/DT, 100.0), -100.0)
        self._last_error = _error       # save for next update
        return _control                 # control output %


class MoveAbsolute:
    def __init__(self, cmd_pos, cmd_vel, cmd_acc, cmd_dec):
        """ Use the constructor to calculate the variables once """
        self._pos = cmd_pos             # command position, mm
        self._vel = cmd_vel             # command velocity, mm/s
        self._acc = cmd_acc             # command acceleration, mm/s^2
        self._dec = cmd_dec             # command deceleration, mm/s^2
        self._t1 = cmd_vel/cmd_acc
        self._x1 = 0.5*cmd_vel*cmd_vel/cmd_acc
        self._t2 = (2.0*cmd_pos*cmd_acc*cmd_dec+cmd_vel**2*(cmd_dec-cmd_acc)) \
                / (2.0*cmd_vel*cmd_acc*cmd_dec)
        self._x2 = (2.0*cmd_pos*cmd_dec-cmd_vel**2)/(2.0*cmd_dec)
        self._t3 = (2*cmd_pos*cmd_acc*cmd_dec+cmd_vel**2*(cmd_acc+cmd_dec)) \
               / (2.0*cmd_vel*cmd_acc*cmd_dec)

    def tg2(self, t):
        """ Second Order Target Generator """
        if t >= 0.0 and t < self._t1:           # accelerating
            _p = self._acc*t*t/2.0
            _v = self._acc*t
            _a = self._acc
        elif t >= self._t1 and t < self._t2:    # constant velocity
            t = t-self._t1
            _p = self._x1+self._vel*t
            _v = self._vel
            _a = 0.0
        elif t >= self._t2 and t < self._t3:    # decelerating
            t = t-self._t2
            _p = self._x2+self._vel*t-self._dec*t*t/2.0
            _v = self._vel - self._dec*t
            _a = self._dec
        elif t >= self._t3:                     # done
            _p = self._pos
            _v = 0.0
            _a = 0.0
        return [_p, _v, _a]

    def get_move_time(self):
        return self._t3


# Initialize the actuator model
K = 10.0                        # open loop gain, mm/s
nf = 2.0*pi*10.0                # natrual frequency, 10 Hz
df = 1.0/3.0                    # damping factor

actuator = Model(K, df, nf)

# Initialize the target generator
# create and instance of a MoveAbsolute command
# move = MoveAbsolute(command_position,                mm
#                     command_velocity,                mm/s
#                     command_acceleration,            mm/s^2
#                     command_deceleration)            mm/s^2
move = MoveAbsolute(200.0, 500.0, 2000.0, 2000.0)
move_time = move.get_move_time()
plot_time = move_time + 0.5     # provide extra time for legend

# Initialize the gains and controller
# When using P only, Kp = 1.258
# When using PI,  Ki = 9.443, Kp = 1.862
# When using PID, Ki = 5.114564, Kp = 5.114564, Kd = 5.114564
Kp = 1.258                      # proportional gain %/mm
Ki = 0.0                        # integrator gain %/(mm*s)
Kd = 0.0                        # derivative gain %/(mm/s)
control = ClosedLoop(Kp, Ki, Kd)

N = int(plot_time/DT)   # number of iterations
tls = np.linspace(0.0, plot_time-DT, N)         # time linear space
r = np.zeros((N, 3))    # target position, velocity and acceleration
x = np.zeros((N, 3))    # actual position, velocity and acceleration
control_output = np.empty_like(tls)

z0 = [0.0, 0.0, 0.0]
for n in range(1, N):
    tspan = [tls[n-1], tls[n]]
    r[n] = move.tg2(tls[n])
    control_output[n] = control.P(r[n-1], x[n-1]) # choose control method
    # control_output[n] = control.PI(r[n-1], x[n-1])
    # control_output[n] = 50.0    # For testing
    z = odeint(actuator.update, z0, tspan, args=(control_output[n],))
    x[n] = z[1]
    z0 = z[1]


# r is the target position, velocity and acceleration array
# x is the actual position, velocity and acceleration array
target_position = r[:, 0]
target_velocity = r[:, 1]
target_acceleration = r[:, 2]
actual_position = x[:, 0]
actual_velocity = x[:, 1]
actual_acceleration = x[:, 2]


# plot position, velocity and control output as a function of time
def make_patch_spines_invisible(ax):
    ax.set_frame_on(True)
    ax.patch.set_visible(False)
    for sp in ax.spines.values():
        sp.set_visible(False)


fig, host = plt.subplots()
fig.set_size_inches(12.0, 8.0)
dpi = fig.get_dpi()
fig.subplots_adjust(right=0.75)

par1 = host.twinx()
par2 = host.twinx()
par2.spines["right"].set_position(("axes", 1.1))
make_patch_spines_invisible(par2)
par2.spines["right"].set_visible(True)

p1, = host.plot(tls, target_position, 'c-', label='target position, mm')
p1a, = host.plot(tls, actual_position, 'r--', label='actual position, mm')
p2, = par1.plot(tls, control_output, 'g-', label='control %')
p3, = par2.plot(tls, target_velocity, 'm-', label='target velocity, mm/s')
p3a, = par2.plot(tls, actual_velocity, 'b--', label='actual velocity, mm/s')

host.set_xlim(0, plot_time)
host.set_ylim(min(actual_position), max(actual_position))
par1.set_ylim(-100*1.1, 100*1.1)
par2.set_ylim(min(actual_velocity), max(actual_velocity))

host.set_title('Position, Velocity and Control vs Time')
host.set_xlabel('time, s')
host.set_ylabel('target and actual positions, mm')
par1.set_ylabel('control %')
par2.set_ylabel('target and actual velocity, mm/s')

# set the color of the text in the legend
host.yaxis.label.set_color(p1.get_color())
# host.yaxis.label.set_color(p1a.get_color())
par1.yaxis.label.set_color(p2.get_color())
par2.yaxis.label.set_color(p3.get_color())
# par2.yaxis.label.set_color(p3a.get_color())

tkw = dict(size=4, width=1.5)
host.tick_params(axis='y', colors=p1.get_color(), **tkw)
par1.tick_params(axis='y', colors=p2.get_color(), **tkw)
par2.tick_params(axis='y', colors=p3.get_color(), **tkw)
host.tick_params(axis='x', **tkw)

lines = [p1, p1a, p2, p3, p3a]
host.legend(lines, [l.get_label() for l in lines], loc=4)

plt.show()
